/* DRFONT expander by John Elliott <jce@seasip.demon.co.uk>, July 2006. For */
/* use in FreeDOS MODE. You may use this code under the terms of the GNU    */
/* GPL (v2) licence if you want to, see www.gnu.org ... */

/* Given a DRFONT CPI and a list of codepages, this expander produces       */
/* the memory image of a standard FONT CPI containing those codepages.      */

#include "mode.h"
#if EXEFILE
#define farmalloc malloc
#define farfree free
#else
#include <alloc.h>
#endif


typedef unsigned char BYTE;
typedef unsigned      WORD;
typedef unsigned long DWORD;

/* CPI file structures */
typedef struct
{
	WORD size;
	DWORD next;
	WORD dev;
	BYTE devname[8];
	WORD codepage;
	BYTE rsvd[6];
	DWORD data;
} CPENTRY;

typedef struct
{
	WORD version;
	WORD count;
	WORD size;
} CPINFO;

typedef struct
{
	BYTE height;
	BYTE width;
	BYTE yaspect;
	BYTE xaspect;
	WORD count;
} SCREENFONT;

/* CPI file values: */
static DWORD st_fih_offset;	/* Offset of codepage count in file */
static WORD st_sizecount;	/* Number of font sizes in file */
static BYTE *st_heights;	/* List of font sizes */
static DWORD *st_offsets;	/* Offsets to their respective tables */
static WORD st_cpcount;		/* Number of codepages in the file */
static DWORD st_wrpos;		/* Calculated font write position */

/* The DRFONT expander was originally written to work on files. Here, it's
 * been redone to use memory buffers. This is why all the reading and 
 * writing is done by helper functions rather than directly accessing the
 * buffers. */

static BYTE far *st_outbuf;
static BYTE far *st_inbuf;
static WORD st_rdptr;
static DWORD st_wrptr;

static void rdseek(DWORD x)
{
	st_rdptr = (WORD)x;
}
#define rdbyte()  (st_inbuf[st_rdptr++])


static WORD rdword(void)
{
	WORD far *p = (WORD far *)(st_inbuf + st_rdptr);
	st_rdptr += 2;
	return *p;
}


static DWORD rdlong(void)
{
	DWORD far *p = (DWORD far *)(st_inbuf + st_rdptr);
	st_rdptr += 4;
	return *p;
}


static void wrbyte(BYTE bt)
{
	if (st_outbuf) st_outbuf[(WORD)st_wrptr] = bt;
	++st_wrptr;
}


static void wrword(WORD w)
{
	if (st_outbuf) 
	{
		WORD far *p = (WORD far *)(st_outbuf + (WORD)st_wrptr);
		*p = w;
	}
	st_wrptr += 2;
}

static void wrlong(DWORD d)
{
	if (st_outbuf) 
	{
		DWORD far *p = (DWORD far *)(st_outbuf + (WORD)st_wrptr);
		*p = d;
	}
	st_wrptr += 4;
}


/* Portable (slower) implementations of the above functions
static WORD rdword()
{
	WORD l = rdbyte();
	WORD h = rdbyte();
	return (h << 8) | l;
}

static DWORD rdlong()
{
	DWORD l = rdword();
	DWORD h = rdword();
	return (h << 16) | l;	
}
*/
static void rdbuf(BYTE *buf, int count)
{
	int n;
	for (n = 0; n < count; n++) buf[n] = rdbyte();
}


static void wrbuf(BYTE *buf, int count)
{
	int n;
	for (n = 0; n < count; n++) wrbyte(buf[n]);
}

/*
static void wrword(WORD w)
{
	wrbyte(w & 0xFF);
	wrbyte(w >> 8);
}

static void wrlong(DWORD l)
{
	wrbyte(l & 0xFF);
	wrbyte(l >> 8);
	wrbyte(l >> 16);
	wrbyte(l >> 24);
}
*/

static void read_cpentry(CPENTRY *e)
{
	e->size = rdword();
	e->next = rdlong();
	e->dev  = rdword();
	rdbuf(e->devname, 8);
	e->codepage = rdword();
	rdbuf(e->rsvd, 6);
	e->data = rdlong();
}

static void read_cpinfo(CPINFO *e)
{
	e->version = rdword();
	e->count   = rdword();
	e->size    = rdword();
}

static void read_scrfont(SCREENFONT *e)
{
	e->height  = rdbyte();
	e->width   = rdbyte();
	e->yaspect = rdbyte();
	e->xaspect = rdbyte();
	e->count   = rdword();
}




static int check_header(void)
{
	BYTE buf[8];
	int n;

	rdseek(0);
	rdbuf(buf, 8);			/* 8 bytes magic */
	if (memcmp(buf, "\177DRFONT ", 8))  return -1;	
	rdbuf(buf, 8);			/* 8 reserved bytes */
	if (rdword() != 1) return -1;	/* Must be 1 pointer */
	if (rdbyte() != 1) return -1;	/* Pointer type must be 1 */
	st_fih_offset = rdlong();
/* Now for the DRFONT extended header */
	st_sizecount = rdbyte();
	st_offsets = malloc(st_sizecount * sizeof(*st_offsets));
	st_heights = malloc(st_sizecount * sizeof(*st_heights));
	if (!st_heights || !st_offsets) return -1;
	for (n = 0; n < st_sizecount; n++)
	{
		st_heights[n] = rdbyte();	
	}	
	for (n = 0; n < st_sizecount; n++)
	{
		st_offsets[n] = rdlong();	
	}
	rdseek(st_fih_offset);
	st_cpcount = rdword();	
	return 0;	
}

/* Get the number of bytes a codepage will occupy in FONT format */
static DWORD entrysize(CPENTRY *cpe)
{
	CPINFO info;
	SCREENFONT sf;
	WORD fontsize;
	WORD n;
	DWORD tot_size, pos;

	rdseek(cpe->data);
	read_cpinfo(&info);
	tot_size = 6;

	pos = cpe->data + 6;
	for (n = 0; n < info.count; n++)
	{
		rdseek(pos);
		read_scrfont(&sf);
		pos += 6;
		tot_size += 6;

		fontsize = (((sf.width+7)/8)*sf.height)*sf.count;
		if (info.version == 1)	/* FONT */
		{
			pos += fontsize;
		}
		tot_size += fontsize;
	}
	return tot_size;
}

static void wrglyph(WORD height, WORD chr)
{
	WORD n, m;
	
	for (n = 0; n < st_sizecount; n++)
	{
		if (st_heights[n] == height)
		{
			rdseek(st_offsets[n] + chr * height);
			for (m = 0; m < height; m++)
				wrbyte(rdbyte());
			return;
		}
	}
	/* Nothing found, fill with zeroes */
	for (m = 0; m < height; m++)
		wrbyte(0);
}


/* Convert a codepage */
static void convert_page(CPENTRY *cpe)
{
	DWORD datasize, rdpos, chartbl;
	WORD n, m, fontsize, chidx;
	CPINFO info;
	SCREENFONT sf;

	datasize = entrysize(cpe);
/* Write entry header */
	wrword(28);			/* Size of header */
	wrlong(st_wrpos + 28 + datasize);	/* Next header */
	wrword(cpe->dev);
	wrbuf(cpe->devname, 8);
	wrword(cpe->codepage);
	wrbuf(cpe->rsvd, 6);
	wrlong(st_wrpos + 28);		/* Offset to data */
	st_wrpos += 28;
/* Write entry data */
	rdseek(cpe->data);
	read_cpinfo(&info);

	wrword(1);			/* FONT */
	wrword(info.count);		/* Number of fonts */
	wrword((WORD)(datasize - 6));	
	rdpos = cpe->data + 6;
	chartbl = cpe->data + 6 + (6 * info.count);
	for (n = 0; n < info.count; n++)
	{
		rdseek(rdpos);
		read_scrfont(&sf);
		wrbyte(sf.height);	/* Screen font header */
		wrbyte(sf.width);
		wrbyte(sf.yaspect);
		wrbyte(sf.xaspect);
		wrword(sf.count);	
		rdpos += 6;
		fontsize = (((sf.width+7)/8)*sf.height)*sf.count;
		if (info.version == 1)	/* FONT font, easy */
		{
			for (m = 0; m < fontsize; m++)
				wrbyte(rdbyte());
			rdpos += fontsize;
		}
		else	/* DRFONT font, not so easy */
		{
			for (m = 0; m < sf.count; m++)
			{
				rdseek(chartbl + 2*m);
				chidx = rdword();
				wrglyph(sf.height, chidx);
			}
		}
	}
	st_wrpos += datasize;
}


/* The actual expansion engine */
int ExpandDRFont(WORD count, WORD *pages, unsigned *length, 
		BYTE far *buf, WORD max)
{
	WORD matches = 0;
	WORD n, m, pass;
	CPENTRY cpe;

	st_inbuf = buf;
	rdseek(0);
	if (check_header()) return -1;
/* Check that we have matches for all requested pages */
	if (count && pages)
	{ 
		rdseek(st_fih_offset + 2);
		for (n = 0; n < st_cpcount; n++)
		{
			read_cpentry(&cpe);
			for (m = 0; m < count; m++)
			{
				if (cpe.codepage == pages[m] && 
						cpe.dev == 1)
					++matches;
			}
			rdseek(cpe.next);
		}
/* We should now have matched all codepages. If we haven't bail out. */
		if (matches < count) 
		{
			free(st_heights); st_heights = NULL;
			free(st_offsets); st_offsets = NULL;
			return -2;
		}
	}
/* Make two passes. The first to calculate how big our expanded FONT 
 * would be, and the second to write it. */
	for (pass = 0; pass < 2; pass++)
	{
		if (pass == 0)
		{
			st_outbuf = NULL;
		}
		else
		{
/* If the DRFONT would expand to bigger than the buffer, bail. */
			if (st_wrptr > max)
			{
				free(st_heights); st_heights = NULL;
				free(st_offsets); st_offsets = NULL;
				return -1;	
			}
/* Create a temporary buffer for the output FONT (it can't expand in-place) */
			st_outbuf = farmalloc(st_wrptr);
			if (st_outbuf == NULL)
			{
				puts("Out of memory");
				exit(1);
			}
		}
		st_wrptr = 0;
/* Start generating the output: Font header */
		wrbyte(0xFF);
		wrbuf((BYTE *)"FONT   \0\0\0\0\0\0\0\0", 15);
		wrword(1);
		wrbyte(1);
		wrlong(23);
/* Codepage count */
		wrword(count);
		st_wrpos = 25;
/* Codepages */
		rdseek(st_fih_offset + 2);
		for (n = 0; n < st_cpcount; n++)
		{
			matches = 0;
			read_cpentry(&cpe);
			if (!count || !pages) matches = 1;
			else for (m = 0; m < count; m++)
			{
				if (cpe.codepage == pages[m] && cpe.dev == 1) 
				{
					matches = 1;
					break;
				}
			}
			if (matches)
			{
				convert_page(&cpe);
			}
			rdseek(cpe.next);
		}
	}
/* Copy the generated FONT back into the passed buffer */
	for (n = 0; n < st_wrptr; n++)
	{
		buf[n] = st_outbuf[n];
	}
	*length = (WORD)st_wrptr;
	farfree(st_outbuf);
	free(st_heights); st_heights = NULL;
	free(st_offsets); st_offsets = NULL;
	return 0;
}

